structure Syntax =
struct

open PPUtil

(* ---------------------------------------------------------------------- *)
(* Datatypes *)

datatype term
  = TmVar of int * int
  | TmAbs of string * term
  | TmApp of term * term

datatype binding = NameBind 

type context = (string * binding) list

datatype command
  = Eval of term
  | Bind of string * binding

fun error msg = 
    (open_box 0;
     print_string "Error: ";
     print_string msg;
     print_newline ();
     close_box ())

(* ---------------------------------------------------------------------- *)
(* Context management *)

val emptycontext = nil

fun ctxlength ctx = List.length ctx

fun addbinding (ctx,x,bind) = (x,bind)::ctx

fun addname (ctx,x) = addbinding(ctx,x,NameBind)

fun isnamebound (ctx,x) = List.exists (fn (y,_) => x = y) ctx

fun pickfreshname (ctx,x) =
    if isnamebound (ctx,x) then pickfreshname (ctx,(x^"'"))
    else ((x,NameBind)::ctx, x)

fun index2name (ctx: context,n) =
    #1(List.nth(ctx,n))
    handle Size =>
      (error (concat["Variable lookup failure: offset: ", Int.toString n,
                     ", context size: ", Int.toString(List.length ctx)]);
       "bogus")

fun name2index (ctx,x) =
    case ctx
      of [] => (error ("Identifier " ^ x ^ " is unbound"); ~1)
       | (y,_)::rest =>
          if y=x then 0
          else 1 + (name2index(rest,x))

(* ---------------------------------------------------------------------- *)
(* Shifting *)

fun tmmap onvar c t = 
    let fun walk(c,t) =
            case t
	      of TmVar(x,n) => onvar(c, x, n)
               | TmAbs(x,t2) => TmAbs(x,walk(c+1,t2))
	       | TmApp(t1,t2) => TmApp(walk(c,t1),walk(c,t2))
     in walk(c,t)
    end

fun termShiftAbove d c t =
    tmmap
      (fn (c,x,n) => if x>=c then TmVar(x+d,n+d) else TmVar(x,n+d))
      c t

fun termShift d t = termShiftAbove d 0 t

(* ---------------------------------------------------------------------- *)
(* Substitution *)

fun termSubst j s t =
    tmmap
     (fn (c, x, n) => if x=j+c then termShift c s else TmVar(x,n))
     0 t

fun termSubstTop s t = 
    termShift (~1) (termSubst 0 (termShift 1 s) t)

 
(* ---------------------------------------------------------------------- *)
(* Printing *)

(* The printing functions call these utility functions to insert grouping
  information and line-breaking hints for the pretty-printing library:
     obox   Open a "box" whose contents will be indented by two spaces if
            the whole box cannot fit on the current line
     obox0  Same but indent continuation lines to the same column as the
            beginning of the box rather than 2 more columns to the right
     cbox   Close the current box
     break  Insert a breakpoint indicating where the line maybe broken if
            necessary.
  See the documentation for the Format module in the OCaml library for
  more details. 
*)

fun obox0() = open_hvbox 0
fun obox() = open_hvbox 2
fun cbox() = close_box()
fun break() = print_break 0 0
val pr = print_string

fun small t = 
    case t
      of TmVar(_,_) => true
       | _ => false

fun printtm_Term(outer, ctx, t) = 
    case t
      of TmAbs(x,t2) =>
         let val (ctx',x') = pickfreshname(ctx, x)
          in obox(); pr "lambda "; pr x'; pr ".";
             if (small t2) andalso not outer then break() else print_space();
             printtm_Term(outer, ctx', t2);
             cbox()
         end
       | t => printtm_AppTerm(outer, ctx, t)

and printtm_AppTerm(outer, ctx, t) =
    case t
      of TmApp(t1, t2) =>
          (obox0();
           printtm_AppTerm(false,ctx,t1);
           print_space();
           printtm_ATerm(false,ctx,t2);
           cbox())
       | t => printtm_ATerm(outer,ctx, t)

and printtm_ATerm(outer, ctx, t) = 
    case t
      of TmVar(x,n) =>
         if ctxlength ctx = n
         then pr (index2name(ctx,x))
         else
	    pr (concat["[bad index: ", Int.toString x, "/", Int.toString n,
		" in {", (List.foldl (fn ((x,_),s) => s ^ " " ^ x) "" ctx), " }]"])
       | t => (pr "("; printtm_Term(outer, ctx, t); pr ")")

fun printtm ctx t = printtm_Term(true, ctx, t) 

fun prbinding ctx NameBind = ()

end (* structure Syntax *)
